use std::{
    collections::HashMap,
    ffi::OsString,
    os::mikros::{self, ipc, syscalls, Errno},
    path::PathBuf,
};

use parking_lot::RwLock;

#[derive(Clone, Copy)]
enum ProcessStatus {
    Running,
    Exited(u8),
}

struct Process {
    stdio: [Option<(u64, u64)>; 3],
    cli_args: Vec<OsString>,
    status: ProcessStatus,
    parent: Option<u64>,
    children: Vec<u64>,
    cwd: PathBuf,
}

struct Serv {
    processes: RwLock<HashMap<u64, Process>>,
}

impl proc_man_rpc::Server for Serv {
    fn new_proc(&self, pid: u64, parent: Option<u64>) -> Result<(), Errno> {
        self.processes.write().insert(
            pid,
            Process {
                stdio: [None; 3],
                cli_args: Vec::new(),
                status: ProcessStatus::Running,
                parent,
                children: Vec::new(),
                cwd: PathBuf::from("/"),
            },
        );
        if let Some(parent) = parent {
            self.processes
                .write()
                .get_mut(&parent)
                .unwrap()
                .children
                .push(pid);
        };
        Ok(())
    }

    fn set_stdio(&self, pid: u64, stdio: [Option<(u64, u64)>; 3]) -> Result<(), Errno> {
        self.processes
            .write()
            .get_mut(&pid)
            .ok_or(Errno::ESRCH)?
            .stdio = stdio;
        Ok(())
    }

    fn get_stdio(&self, from: u64) -> [Option<(u64, u64)>; 3] {
        self.processes
            .read()
            .get(&from)
            .map_or([None; 3], |proc| proc.stdio)
    }

    fn set_cli_args(&self, pid: u64, args: Vec<OsString>) -> Result<(), Errno> {
        self.processes
            .write()
            .get_mut(&pid)
            .ok_or(Errno::ESRCH)?
            .cli_args = args;
        Ok(())
    }

    fn get_cli_args(&self, from: u64) -> Vec<OsString> {
        self.processes
            .read()
            .get(&from)
            .map_or(Vec::new(), |proc| proc.cli_args.clone())
    }

    fn exit(&self, pid: u64, code: u8) {
        if let Some(proc) = self.processes.write().get_mut(&pid) {
            proc.status = ProcessStatus::Exited(code);
        }
    }

    fn wait(&self, from: u64, pid: u64, block: bool) -> Result<proc_man_rpc::WaitResult, Errno> {
        let processes = self.processes.read();
        let proc = processes.get(&from).ok_or(Errno::ESRCH)?;
        if pid != 0 && !proc.children.contains(&pid) {
            return Err(Errno::ECHILD);
        }
        std::mem::drop(processes);
        if pid == 0 {
            loop {
                let processes = self.processes.read();
                let proc = processes.get(&from).ok_or(Errno::ESRCH)?;
                for &pid in &proc.children {
                    let proc_status = self.processes.read().get(&pid).unwrap().status;
                    if let ProcessStatus::Exited(code) = proc_status {
                        self.processes.write().remove(&pid);
                        syscalls::clear_exited(pid).unwrap();
                        self.processes
                            .write()
                            .get_mut(&from)
                            .unwrap()
                            .children
                            .retain(|&child| child != pid);
                        return Ok(proc_man_rpc::WaitResult {
                            pid,
                            exit_code: code,
                        });
                    };
                }
                std::mem::drop(processes);
                if !block {
                    return Err(Errno::EAGAIN);
                }
                ipc::process_messages();
            }
        } else {
            loop {
                let proc_status = self.processes.read().get(&pid).unwrap().status;
                if let ProcessStatus::Exited(code) = proc_status {
                    self.processes.write().remove(&pid);
                    syscalls::clear_exited(pid).unwrap();
                    self.processes
                        .write()
                        .get_mut(&from)
                        .unwrap()
                        .children
                        .retain(|&child| child != pid);
                    return Ok(proc_man_rpc::WaitResult {
                        pid,
                        exit_code: code,
                    });
                };
                if !block {
                    return Err(Errno::EAGAIN);
                }
                ipc::process_messages();
            }
        }
    }

    fn get_processes(&self) -> Vec<proc_man_structs::Process> {
        let mut procs = self
            .processes
            .read()
            .iter()
            .map(|(&pid, proc)| {
                let status = match proc.status {
                    ProcessStatus::Running => proc_man_structs::ProcessStatus::Running,
                    ProcessStatus::Exited(code) => proc_man_structs::ProcessStatus::Exited(code),
                };
                proc_man_structs::Process {
                    pid,
                    status,
                    stdio: proc.stdio,
                    cli_args: proc.cli_args.clone(),
                    parent: proc.parent,
                    children: proc.children.clone(),
                }
            })
            .collect::<Vec<_>>();
        procs.sort_by_key(|proc| proc.pid);
        procs
    }

    fn set_cwd(&self, pid: u64, path: PathBuf) -> Result<(), Errno> {
        if !path.is_absolute() {
            return Err(Errno::EINVAL);
        }
        // Try to open the given directory, if it fails it's not a valid directory.
        // Use Metadata instead once it's implemented.
        std::fs::read_dir(&path)?;
        self.processes
            .write()
            .get_mut(&pid)
            .ok_or(Errno::ESRCH)?
            .cwd = path;
        Ok(())
    }

    fn get_cwd(&self, from: u64) -> PathBuf {
        self.processes
            .read()
            .get(&from)
            .map_or(PathBuf::from("/"), |proc| proc.cwd.clone())
    }
}

fn main() {
    let syslog_pid = loop {
        if let Some(pid) = syscalls::try_get_registered(2) {
            break pid;
        }
    };
    let syslog_client = syslog_rpc::Client::new(syslog_pid);
    let serv = Serv {
        processes: RwLock::new(HashMap::new()),
    };
    proc_man_rpc::register_server(Box::new(serv));
    syscalls::register(3);
    syslog_client
        .send_text_message("proc_man", "Process manager initialized")
        .unwrap();
    loop {
        ipc::process_messages();
    }
}
